;;;
;;; scbib - a bibliography management system written in Scheme
;;;
;;; Copyright (C) 2004 Satoru Takabayashi <satoru@namazu.org> 
;;;     All rights reserved.
;;;     This is free software with ABSOLUTELY NO WARRANTY.
;;;
;;; You can redistribute it and/or modify it under the terms of 
;;; the Gauche's licence.
;;;
(define-module scbib
  (use srfi-1)
  (use gauche.parameter)
  (use file.util)

  (export scbib-db-person
          scbib-db-publisher
          scbib-db-journal
          scbib-db-bib
          scbib-find
          scbib-values
          scbib-value
          scbib-load-db
          scbib-load-db-from-port
          scbib-load-db-all
          scbib-load-path
          scbib-match?
          scbib-get-authors
          scbib-get-abbrev
          scbib-print-abbrev-list
	  ))
(select-module scbib)

;; DBs
(define scbib-db-person    (make-parameter '()))
(define scbib-db-publisher (make-parameter '()))
(define scbib-db-journal   (make-parameter '()))
(define scbib-db-bib       (make-parameter '()))

(define scbib-load-path (make-parameter "."))

(define (scbib-find query db)
  (cond ((null? db)
	 #f)
	((query (car db))
	 (car db))
	(else 
	 (scbib-find query (cdr db)))))

(define (scbib-values item name)
  (cond ((null? item)
	 #f)
	((eq? (caar item) name)
	 (cdar item))
	(else
	 (scbib-values (cdr item) name))))

(define (scbib-value item name)
  (let ((val (scbib-values item name)))
    (if val
	(car val)
	#f)))

(define (scbib-load-db-from-port iport)
  (define (add-to-db! item)
    (let* ((db-type (car item))
	   (contents (cdr item)))
      (case db-type
	((person) 
	 (scbib-db-person
          (cons contents (scbib-db-person))))
	((publisher)
	 (scbib-db-publisher
          (cons contents (scbib-db-publisher))))
	((journal)
	 (scbib-db-journal
          (cons contents (scbib-db-journal))))
	((bib)
	 (scbib-db-bib
          (cons contents (scbib-db-bib)))))))
  (let loop ()
    (let ((item (read iport)))
      (unless (eof-object? item)
              (add-to-db! item)
              (loop)))))

(define (scbib-load-db file)
  (let1 file (if (file-exists? file)
                 file
                 (build-path (expand-path (scbib-load-path))
                             file))
        (call-with-input-file file
          (lambda (iport)
            (scbib-load-db-from-port iport)))))

(define (scbib-load-db-all)
  (for-each (lambda (file)
              (scbib-load-db file))
            (directory-list (expand-path (scbib-load-path))
                            :add-path? #t
                            :children? #t
                            :filter #/\.db$/)))

;; (scbib-match item :bibtype "book")
;; (scbib-match item :bibtype "book" :category '(not "programming"))
;; (scbib-match item :bibtype '(or "book" "article"))
(define (scbib-match? item  . options)
  (let-keywords* options 
    ((bibtype #f)
     (category #f)
     (subcategory #f))
    (every (lambda (key value)
             (or (eq? value #f)
                 (let1 v (scbib-value item key)
                       (cond ((and (pair? value) (eq? (first value) 'not))
                              (not (equal? v (second value))))
                             ((and (pair? value) (eq? (first value) 'or))
                              (any (lambda (vv) (equal? v vv))
                                   (cdr value)))
                             (else
                              (equal? v value))))
                 ))
           '(bibtype category subcategory)
           (list bibtype category subcategory))))



(define scbib-get-abbrev
  (let ((abbrev-alist '()))
    (lambda (bibitem)
      (define (family-name name)
        ;; Japanese name
        (let ((m (rxmatch #/^([^a-z ]+) [^a-z ]+$/ name)))
          (if m
              (rxmatch-substring m 1)
              (let ((m (rxmatch #/([^ ]+)( Jr\.)$/ name)))
                (if m
                    (rxmatch-substring m 1)
                    (let ((m (rxmatch #/([^ ]+)$/ name)))
                      (rxmatch-substring m 1)
                      ))))))

      (define (generate-abbrev)
        (let* ((authors (scbib-get-authors bibitem))
               (author (if authors (car authors) #f))
               (editor (scbib-value bibitem 'editor)))
          (let ((author (family-name (or author
                                         editor
                                         "Anonymous")))
                (year (or (scbib-value bibitem 'year) "XXXX")))
            (format #f "~a~a" author year))))

      (let* ((abbrev-candidate
              (or (scbib-value bibitem 'abbrev) (generate-abbrev)))
             (abbrev-item (assoc abbrev-candidate abbrev-alist)))
        (if abbrev-item 
            (let ((count (cdr abbrev-item)))
              (set-cdr! abbrev-item (+ count 1))
              (string-append abbrev-candidate "-" 
                             (number->string (+ count 1))))
            (begin 
              (set! abbrev-alist 
                    (cons (cons abbrev-candidate 1) abbrev-alist))
              abbrev-candidate))))))

(define (scbib-get-authors bibitem)
  (let ((title (scbib-value bibitem 'title))
        (original-title (scbib-value bibitem 'original-title)))
    (if original-title 
        ;; if the original title exists
        (let ((original-bibitem
               (scbib-find (lambda (x) 
                             (equal? (car (scbib-values x 'title)) 
                                     original-title)) 
                           (scbib-db-bib))))
          (append (scbib-values original-bibitem 'author)
                  (let1 r (reverse (scbib-values bibitem 'author))
                        (reverse (cons (string-append (car r) " ")
                                       (cdr r))))))
        (scbib-values bibitem 'author))))

(define (scbib-print-abbrev-list)
  (for-each (lambda (bib) 
	      (format #t "~a\n" (scbib-value bib 'abbrev)))
	    (reverse (scbib-db-bib))))
