;;;
;;; scbib - a bibliography management system written in Scheme
;;;
;;; Copyright (C) 2004 Satoru Takabayashi <satoru@namazu.org> 
;;;     All rights reserved.
;;;     This is free software with ABSOLUTELY NO WARRANTY.
;;;
;;; You can redistribute it and/or modify it under the terms of 
;;; the Gauche's licence.
;;;

(define-module scbib.html
  (use gauche.regexp)
  (use gauche.parameter)
  (use srfi-1)
  (use scbib)
  (use scbib.amazon)
  (use rfc.uri)
  (export scbib-html-print
          scbib-html-print-items
          scbib-html-title-first?
          ))
(select-module scbib.html)


(define (escape string)
  (set! string (regexp-replace-all #/[{}]/ string ""))
  (set! string (regexp-replace-all #/&/ string "&amp;"))
  (set! string (regexp-replace-all #/</ string "&lt;"))
  (set! string (regexp-replace-all #/>/ string "&gt;"))
  string)


(define scbib-html-title-first? (make-parameter #f))

(define (link-keyword string db)
  (regexp-replace-all #/\{([^{}]+)\}/ 
                      string 
                      (lambda (m)
                        (link2 (rxmatch-substring m 1)
                               db
                               '(title abbrev)))))

(define (link-format string url)
  (string-append "<a href=\"" 
                 (escape url)
                 "\">" 
                 (escape string)
                 (if (rxmatch #/\.pdf$/ url)
                     " (PDF)"
                     "")
                 "</a>"))

(define (link string db key)
  (let ((item (scbib-find (lambda (x) 
                               (equal? (scbib-value x key)
                                       string))
                             db)))
    (if item
        (let ((web (scbib-values item 'web)))
          (if web
              (link-format string (car web))
              string))
        string)))

(define (link2 string db keys)
  (cond ((null? keys)
         string)
        (else
         (let ((linked (link string db (car keys))))
           (if (equal? linked string)
               (link2 string db (cdr keys))
               linked)))))

(define (get-web string db key)
  (let ((item (scbib-find (lambda (x) 
                               (equal? (scbib-value x key)
                                       string))
                             db)))
    (if item
        (or (scbib-value item 'web) #f)
        #f)))

(define (get-web2 string db keys)
  (cond ((null? keys)
         string)
        (else
         (or (get-web string db (car keys))
             (get-web2 string db (cdr keys))))))


(define (scbib-html-print bibdb bibitem lang)
  (define (scbib-html-generate bibitem)
    (define (editors)
      (let ((editors (scbib-values bibitem 'editor))
            (suffix (if (equal? lang "ja")
                        " "
                        ", editor")))
        (cond ((eq? editors #f)
               #f)
              ((= (length editors) 1)
               (string-append (link (car editors) (scbib-db-person) 'name) suffix))
              (else
               (string-append (string-join
                               (map (lambda (editor)
                                      (link editor (scbib-db-person) 'name))
                                    editors) 
                               ", ")
                              suffix)))))
    
    (define (link-author author)
      (let1 m (rxmatch #/^(.*)( )$/ author)
            (if m
                (string-append (link (rxmatch-substring m 1)
                                     (scbib-db-person) 'name)
                               (rxmatch-substring m 2))
                (link author (scbib-db-person) 'name))))

    (define (authors)
      (let ((authors (scbib-get-authors bibitem)))
        (cond ((eq? authors #f)
               (editors))
              ((= (length authors) 1)
               (link (car authors) (scbib-db-person) 'name))
              (else
               (string-join (map (lambda (author)
                                   (link-author author))
                                 authors) ", ")))))

    (define (web)
      (escape (or (scbib-value bibitem 'web) "")))

    (define (left-paren)
      (if (and (equal? lang "ja")
               (scbib-html-title-first?))
          ""
          ""))

    (define (right-paren)
      (if (and (equal? lang "ja")
               (scbib-html-title-first?))
          ""
          ""))

    (define (title)
      (string-append 
       "<cite>"
       (left-paren)
       (cond ((scbib-value bibitem 'isbn)
              (link-format 
               (scbib-value bibitem 'title)
               (scbib-amazon-web-url
                (scbib-value bibitem 'isbn))))
             ((scbib-value bibitem 'web)
              (link-format 
               (scbib-value bibitem 'title)
               (scbib-value bibitem 'web)))
             (else (escape (scbib-value bibitem 'title))))
       
       (let ((series (scbib-value bibitem 'series)))
         (if series
             (string-append " (" series ")")
             ""))
       (right-paren)
       "</cite>" 
       ))

    (define (postscript)
      (let ((postscript (scbib-value bibitem 'postscript)))
        (if postscript
            (string-append "<a href=\"" 
                           (escape postscript)
                           "\">" "[PostScript]" "</a>")
            "")))

    (define (pdf)
      (let ((pdf (scbib-value bibitem 'pdf)))
        (if pdf
            (string-append "<a href=\"" 
                           (escape pdf)
                           "\">" "[PDF]" "</a>")
            "")))

    (define (video)
      (let ((video (scbib-value bibitem 'video)))
        (if video
            (string-append "<a href=\"" 
                           (escape video)
                           "\">" 
                           (if (equal? lang "ja") "[ӥǥ]" "Video")
                           "</a>")
            "")))

    (define (handout)
      (let ((handout (scbib-value bibitem 'handout)))
        (if handout
            (string-append "<a href=\"" 
                           (escape handout)
                           "\">" 
                           (if (equal? lang "ja") "[ȯɽ]" "Handout")
                           "</a>")
            "")))

    (define (publisher)
      (link (scbib-value bibitem 'publisher) (scbib-db-publisher) 'name))

    (define (booktitle)
      (scbib-value bibitem 'booktitle))

    (define (journal)
      (link (scbib-value bibitem 'journal) (scbib-db-journal) 'name))

    (define (phdthesis)
      (if (equal? (scbib-value bibitem 'bibtype)
                  "phdthesis")
          "PhD thesis"
          #f))

    (define (mastersthesis)
      (if (equal? (scbib-value bibitem 'bibtype)
                  "mastersthesis")
          "Master's thesis"
          #f))

    (define (year)
      (scbib-value bibitem 'year))

    (define (month)
      (or (scbib-value bibitem 'month) ""))

    (define (event)
      (or (scbib-value bibitem 'event) ""))

    (define (note)
      (or (scbib-value bibitem 'note) ""))

    (define (pages)
      (let ((pages (scbib-value bibitem 'pages)))
        (if pages 
            (string-append 
             (if (rxmatch #/-/ pages)
                 "pp. "
                 "p. ")
             pages)
            #f)))

    (define (year+edition)
      (let ((year (year))
            (biblang (scbib-value bibitem 'lang))
            (edition (scbib-value bibitem 'edition)))
        (if (and year edition)
            (string-append year
                           " (" 
                           (cond ((not biblang) edition)
                                 ((and (equal? biblang "ja")
                                       (rxmatch #/^[0-9]+$/ edition))
                                  (string-append "" edition ""))
                                 ((and (equal? biblang "en")
                                       (rxmatch #/Second/ edition))
                                  "2nd ed.")
                                 ((and (equal? biblang "en")
                                       (rxmatch #/Third/ edition))
                                  "3rd ed.")
                                 ((and (equal? biblang "en")
                                       (rxmatch #/Fourth/ edition))
                                  "4th ed.")
                                 ((equal? biblang "en")
                                  (string-append edition " ed."))
                                 (else edition))
                               ")")
            (or year ""))))

    (define (source)
      (string-join (filter 
                    (lambda (not-false) not-false)
                    (list 
                     (or (journal) (booktitle) (phdthesis) (mastersthesis) #f)
                     (let ((volume (scbib-value bibitem 'volume)))
                       (if volume 
                           (string-append "Vol. " volume)
                           #f))
                     (let ((number (scbib-value bibitem 'number)))
                       (if number 
                           (string-append "No. " number)
                           #f))
                     (or (pages) #f))) ", "))

    (define (generate-book)
      (let* ((first  (if (scbib-html-title-first?) (title) (authors)))
             (second (if (scbib-html-title-first?) (authors) (title)))
             (title+authors (if (authors)
                                (format "~a: ~a, " first second)
                                (format "~a: " (title)))))
        (format "~a~a, ~a." 
                title+authors
                (publisher)
                (year+edition))))

    (define (generate-paper)
      (format "~a: ~a." 
              (authors)
              (string-join (filter
                                 (lambda (x) (not (equal? x "")))
                                 (list (title)
                                       (postscript)
                                       (pdf)
                                       (handout)
                                       (video)
                                       (source)
                                       (event)
                                       (month)
                                       (year))) ", ")))
    (define (generate-web)
      (format "~a~a" 
              (if (authors)
                  (string-append (authors) ": ")
                  "")
              (title)))

    (let ((bibtype (scbib-value bibitem 'bibtype)))
      (cond ((equal? bibtype "book")
             (generate-book))
            ((equal? bibtype "web")
             (generate-web))
            (else
             (generate-paper)))))

  (define (add-authors bibitem author)
    (define (add-authors-internal item result)
      (cond ((null? item)
             result)
            ((eq? (caar item) 'author)
             (add-authors-internal (cdr item)
                                  (cons
                                   (cons 'author (append author (cdar item)))
                                   result)))
            (else
             (add-authors-internal (cdr item) 
                                  (cons (car item) result)))))

    (add-authors-internal bibitem '()))

  (define (delete-authors bibitem)
    (define (delete-authors-internal item result)
      (cond ((null? item)
             result)
            ((eq? (caar item) 'author)
             (delete-authors-internal (cdr item)
                                      result))
            (else
             (delete-authors-internal (cdr item) 
                                      (cons (car item) result)))))

    (delete-authors-internal bibitem '()))

  (define (print-ja-book bibitem)
    (let* ((original-title (scbib-value bibitem 'title))
           (translation 
            (scbib-find (lambda (x) 
                          (equal? (scbib-value x 'original-title)
                                  original-title))
                        bibdb)))
      (if translation
          (begin
            (display (scbib-html-generate translation))
            (format "<em></em>: ~a"
                  (scbib-html-generate (delete-authors bibitem))))
          (begin (display (scbib-html-generate bibitem))
                 #f))))

  (define (get-usage)
    (define (normalize string)
      (set! string (regexp-replace-all #/\n        / string "\n"))
      (set! string (regexp-replace-all #/^\n/ string "")))

    (let ((usage (scbib-value bibitem 'usage)))
      (if usage
          (format "<pre>\n~a\n</pre>" 
                  (escape (normalize (scbib-value bibitem 'usage))))
          #f)))

  (define (get-comment)
    (let ((comment (scbib-value bibitem 'comment)))
      (if comment
          (link-keyword 
           (escape (scbib-value bibitem 'comment))
           bibdb)
          #f)))

  (define (get-info)
    (let ((info (scbib-values bibitem 'info)))
      (if info
          (format "<em></em>: ~a"
                  (string-join
                   (map (lambda (i)
                          (if (pair? i)
                              (format #f "<a href=\"~a\">~a</a>"
                                      (get-web2 (cadr i) bibdb 
                                                '(title abbrev))
                                      (car i))
                              (link2 i bibdb '(abbrev title))))
                        info)
                   ",<br />"))
          #f)))

  (define (print-additional-info additional-info)
    (let1 additional-info (filter (lambda (x) x) additional-info)
          (unless (null? additional-info)
                  (print "\n<ul class='small'>")
                  (for-each 
                   (lambda (x) (format #t "<li>~a</li>\n"
                                       x))
                   (reverse additional-info))
                  (print "</ul>"))))

  (define (print-anchor bibitem)
    (let1 abbrev (regexp-replace-all 
                  #/%/ (uri-encode-string (scbib-get-abbrev bibitem)) "")
          (format #t " <a id='~a' name='~a' href='#~a'>*</a>"
                  abbrev abbrev abbrev)))

  (let ((title (scbib-value bibitem 'title))
        (original-title (scbib-value bibitem 'original-title)))
    
    ;; if the original title exists, skip it.
    (when (not (and original-title 
                    (scbib-find (lambda (x) 
                                  (equal? (car (scbib-values x 'title))
                                          title))
                                bibdb)))
          ;; if lang is "en" and the item's lang is not "en", skip it.
          (unless (and (equal? lang "en")
                       (not (equal? (scbib-value bibitem 'lang) "en")))
                  (let ((additional-info '()))
                    (display "<li>")
                    (if (and (equal? lang "ja")
                             (equal? (scbib-value bibitem 'bibtype) "book"))
                        (push! additional-info (print-ja-book bibitem))
                        (display (scbib-html-generate bibitem)))
                    (print-anchor bibitem)
                    (when (equal? lang "ja")
                          (push! additional-info (get-comment))
                          (push! additional-info (get-info))
                          (push! additional-info (get-usage)))
                    (print-additional-info additional-info)
                    (display "</li>\n")
                    )))))


(define (scbib-html-print-items lang . options)
  (for-each (lambda (item)
              (if (apply scbib-match? item options)
                  (scbib-html-print (scbib-db-bib) item lang)))
            (reverse (scbib-db-bib))))

